VERSION = "0.5.7.1"
"""bioimageio.spec version as MAJOR.MINOR.PATCH.LIB

MAJOR.MINOR.PATCH correspond to the latest model description format version implemented.

LIB may be bumped for library changes that do not affect the format version.
"""
